from fractions import Fraction
from functools import reduce, wraps

import attr
from aiger_bv import identity_gate, UnsignedBVExpr

import aiger_coins as aigc


def unrelated_coins(left, right):
    return len(left.inputs & right.inputs) == 0


def same_coins(left, right):
    return left.valid.aigbv == right.valid.aigbv


def unary_op(op):
    @wraps(op)
    def _unary_op(dice):
        return dice.evolve(expr=op(dice.expr), valid=dice.valid)

    return _unary_op


def binop(op):
    @wraps(op)
    def binop(left, right):
        if isinstance(right, int):
            return left.evolve(expr=op(left.expr, 1), valid=left.valid)

        expr = op(left.expr, right.expr)
        if same_coins(left, right):
            return left.evolve(expr=expr, valid=left.valid)

        assert unrelated_coins(left, right)
        return left.evolve(expr=expr, valid=left.valid & right.valid)

    return binop


@attr.s(frozen=True, auto_attribs=True, eq=False, order=False)
class Distribution:
    expr: UnsignedBVExpr
    valid: UnsignedBVExpr

    @property
    def aig(self):
        return self.aigbv.aig

    @property
    def aigbv(self):
        return self.expr.aigbv

    @property
    def output(self):
        return self.expr.output

    @property
    def inputs(self):
        return self.expr.inputs

    @property
    def size(self):
        return self.expr.size

    def __getitem__(self, idx):
        expr = self.expr[idx]
        kind = Coin if isinstance(idx, int) else Distribution
        return kind(expr=expr, valid=self.valid)

    def apply(self, func):
        expr = func(self.expr)
        kind = Coin if expr.size == 1 else Distribution
        return kind(expr=expr, valid=self.valid)

    @property
    def coins(self) -> UnsignedBVExpr:
        coins = (
            identity_gate(self.aigbv.imap[i].size, i) for i in self.inputs
        )
        coins = reduce(lambda x, y: x | y, coins)
        return UnsignedBVExpr(coins)

    def condition(self, expr):
        return type(self)(self.expr, expr & self.valid)

    def freqs(self):
        return tuple(self[i].prob() for i in range(self.size))

    def evolve(self, **kwargs):
        return attr.evolve(self, **kwargs)

    def concat(self, other):
        dist = self.apply(lambda x: x.concat(other))
        if same_coins(self, other):
            return dist
        assert unrelated_coins(self, other)
        return dist.condition(other.valid)

    def with_output(self, name):
        return attr.evolve(self, expr=self.expr.with_output(name))

    __add__ = binop(UnsignedBVExpr.__add__)
    __le__ = binop(UnsignedBVExpr.__le__)
    __lt__ = binop(UnsignedBVExpr.__lt__)
    __gt__ = binop(UnsignedBVExpr.__gt__)
    __ge__ = binop(UnsignedBVExpr.__ge__)
    __eq__ = binop(UnsignedBVExpr.__eq__)
    __ne__ = binop(UnsignedBVExpr.__ne__)
    __invert__ = unary_op(UnsignedBVExpr.__invert__)
    __or__ = binop(UnsignedBVExpr.__or__)
    __and__ = binop(UnsignedBVExpr.__and__)
    __xor__ = binop(UnsignedBVExpr.__xor__)

    def __rshift__(self, other):
        assert isinstance(other, aigc.MDP)
        return other << self


@attr.s(frozen=True)
class Coin(Distribution):
    expr: UnsignedBVExpr = attr.ib()
    valid: UnsignedBVExpr = attr.ib()

    @expr.validator
    def check_expr(self, attribute, value):
        if value.size != 1:
            raise ValueError("Must be a 1 sided dice.")

    def prob(self):
        import aiger_bdd
        top, bot = map(aiger_bdd.count, (self.expr & self.valid, self.valid))
        return Fraction(top, bot)



import attr
import funcy as fn

import aiger_bv as BV
from aiger_bv import atom, identity_gate, AIGBV
from bidict import bidict
from pyrsistent import pmap
from pyrsistent.typing import PMap

import aiger_coins as aigc


def _create_input2dist(input2dist):
    return pmap({
        k: dist.with_output(k) if k != dist.output else dist
        for k, dist in input2dist.items()
    })


@attr.s(frozen=True, auto_attribs=True, eq=False, order=False)
class MDP:
    _aigbv: AIGBV
    input2dist: PMap[str, aigc.Distribution] = attr.ib(
        default=pmap(), converter=_create_input2dist
    )

    @property
    def env_inputs(self):
        return set(self.input2dist.keys())

    @property
    def inputs(self):
        return self._aigbv.inputs - self.env_inputs

    @property
    def outputs(self):
        return self._aigbv.outputs

    @property
    def aigbv(self):
        assert "##valid" not in self.outputs

        circ = self._aigbv
        is_valid = atom(1, 1, signed=False)
        for dist in self.input2dist.values():
            circ <<= dist.expr.aigbv
            is_valid &= dist.valid

        circ |= is_valid.with_output("##valid").aigbv
        return circ

    @property
    def aig(self):
        return self.aigbv.aig

    def __lshift__(self, other):
        if isinstance(other, aigc.Distribution):
            other = dist2mdp(other)
        elif isinstance(other, AIGBV):
            other = circ2mdp(other)

        assert isinstance(other, MDP)
        assert not (self.env_inputs & other.env_inputs)
        return circ2mdp(
            circ=self._aigbv << other._aigbv,
            input2dist=self.input2dist + other.input2dist,
        )

    def __rshift__(self, other):
        assert not isinstance(other, aigc.Distribution)
        return other << self

    def __or__(self, other):
        assert not (self.env_inputs & other.env_inputs)
        return circ2mdp(
            circ=self._aigbv | other._aigbv,
            input2dist=self.input2dist + other.input2dist,
        )

    def feedback(self, inputs, outputs, initials=None, latches=None,
                 keep_outputs=False):
        assert set(inputs) <= self.inputs
        circ = self._aigbv.feedback(
            inputs, outputs, initials, latches, keep_outputs
        )
        return circ2mdp(circ, input2dist=self.input2dist)

    def encode_trc(self, sys_actions, states):
        coin_flips = find_coin_flips(sys_actions, states, mdp=self)
        return [fn.merge(a, c) for a, c in zip(sys_actions, coin_flips)]

    def decode_trc(self, actions):
        circ = self.aigbv
        sys_actions = [fn.project(a, self.inputs) for a in actions]

        states = fn.lpluck(0, circ.simulate(actions))
        assert all(s['##valid'] for s in states)
        states = [fn.omit(s, {'##valid'}) for s in states]

        return sys_actions, states

    @fn.memoize
    def _reach_bdd(self):
        try:
            from aiger_bdd import to_bdd
        except ImportError:
            raise ImportError(
                "Need to install py-aiger-bdd to use this method."
            )

        circ = self.aigbv
        assert not ((circ.inputs | circ.outputs) & circ.latches)
        step, lmap = (self.aigbv >> BV.sink(1, ['##valid'])).cutlatches()

        for name, _ in lmap.values():
            step >>= BV.sink(step.omap[name].size, [name])

        check_eq = None
        for v in step.outputs:
            tmp = BV.atom(step.omap[v].size, v, signed=False) \
                == BV.atom(step.omap[v].size, f"{v}##next", signed=False)

            if check_eq is None:
                check_eq = tmp
            else:
                check_eq &= tmp

        reachable = step >> check_eq.aigbv
        assert len(reachable.outputs) == 1
        bexpr, _, _ = to_bdd(reachable, renamer=lambda _, x: x)
        prev_latch = dict(lmap.values())
        new2old = {v[0]: k for k, v in lmap.items()}
        return bexpr, prev_latch, reachable.imap, bidict(new2old)


def circ2mdp(circ, input2dist=None):
    if not isinstance(circ, AIGBV):
        circ = circ.aigbv

    if input2dist is None:
        return MDP(circ)

    return MDP(circ, input2dist=input2dist)


def dist2mdp(dist):
    circ = identity_gate(dist.size, dist.output)
    return circ2mdp(circ=circ, input2dist={dist.output: dist})


def find_coin_flips(actions, states, mdp):
    return list(_find_coin_flips(actions, states, mdp))


def _constraint(k, v):
    var = atom(len(v), k, signed=False)
    return var == BV.decode_int(v, signed=False)


def _find_coin_flips(actions, states, mdp):
    assert len(actions) == len(states)
    circ1 = mdp.aigbv
    bexpr, prev_latch, imap, relabels = mdp._reach_bdd()

    for action, state in zip(actions, states):
        state = fn.walk_keys("{}##next".format, state)

        assignment = fn.merge(prev_latch, action, state)
        coin_flips = bexpr.let(**imap.blast(assignment)).pick()
        coin_flips = imap.project(mdp.env_inputs).unblast(coin_flips)

        yield coin_flips

        if len(prev_latch) > 0:
            ext_action = fn.merge(action, coin_flips)

            latch_vals = fn.walk_keys(relabels.get, prev_latch)
            next_latch = circ1(ext_action, latches=latch_vals)[1]
            assert next_latch.keys() == latch_vals.keys()
            prev_latch = fn.walk_keys(relabels.inv.get, next_latch)



from fractions import Fraction
import hypothesis.strategies as st
from hypothesis import given, settings

import aiger_bv

from aiger_coins import binomial, coin, dist


@settings(max_examples=4, deadline=None)
@given(st.integers(0, 7), st.integers(1, 7))
def test_biased_coin(k, m):
    f = Fraction(*sorted([k, m]))
    k, m = f.numerator, f.denominator
    for prob in [f, (k, m)]:
        assert coin((k, m), 'x').prob() == f


@settings(max_examples=4, deadline=None)
@given(st.lists(st.integers(1, 5), min_size=2, max_size=10))
def test_dist(weights):
    denom = sum(weights)
    freqs = [Fraction(w, denom) for w in weights]

    mux = dist(freqs)

    def is_one_hot(x):
        x = aiger_bv.SignedBVExpr(x.aigbv)
        test = (x != 0) & ((x & (x - 1)) == 0)
        return aiger_bv.UnsignedBVExpr(test.aigbv)

    one_hot = mux.apply(is_one_hot)[0]
    assert one_hot.prob() == 1

    for f1, f2 in zip(freqs, mux.freqs()):
        assert f1 == f2


def test_dice():
    weights = [1, 3, 2]
    freqs = [Fraction(w, 6) for w in weights]
    dice = dist(freqs)

    assert dice.freqs() == tuple(freqs)

    assert (dice[0] | dice[1]).prob() == Fraction(2, 3)
    assert (~dice[2]).prob() == Fraction(2, 3)

    dice2 = dice.apply(lambda expr: ~expr[2])
    assert dice2.freqs() == (Fraction(2, 3),)

    assert len(dice.inputs) == 1
    assert len(dice.aigbv.outputs) == 1
    assert dice.output == dice.expr.output

    assert len(dice.aig.inputs) == 3
    assert len(dice.aig.outputs) == dice.size

    dice3 = dice.condition(dice.coins != 0)
    new_freqs = [Fraction(w, 5) for w in [0, 3, 2]]
    assert dice3.freqs() == tuple(new_freqs)

    dice4 = dice[0].concat(dice[1]).concat(dice[2])
    assert dice4.freqs() == tuple(freqs)


def test_binomial():
    weights = [1, 6, 15, 20, 15, 6, 1]
    expected_freqs = [Fraction(v, 64) for v in weights]

    assert binomial(6).freqs() == tuple(expected_freqs)



import math
from fractions import Fraction

from aiger_bv import atom, encode_int, ite


# Python is deprecating fractions.gcd....
# https://stackoverflow.com/questions/147515/least-common-multiple-for-3-or-more-numbers/147539#147539  # noqa: E501
def lcm(a, b):
    """Return lowest common multiple."""
    return a * b // math.gcd(a, b)


# Simplification of py-aiger-gridworld's chain.
def chain(n, state_name='x', action='H'):
    bits = n + 1
    start = encode_int(bits, 1, signed=False)

    x = atom(bits, state_name, signed=False)
    forward = atom(1, action, signed=False)

    x2 = ite(forward, x << 1, x)
    circ = x2.aigbv['o', {x2.output: state_name}]
    return circ.feedback(
        inputs=[state_name], outputs=[state_name],
        latches=[f"{state_name}_prev"],
        initials=[start], keep_outputs=True
    )


def to_frac(prob):
    prob = Fraction(*prob) if isinstance(prob, tuple) else Fraction(prob)
    assert 0 <= prob <= 1
    return prob



import math
from functools import reduce

import funcy as fn
from aiger_bv import atom, UnsignedBVExpr

from aiger_coins import utils
from aiger_coins.dice import Distribution


def coin(prob, name=None):
    prob = utils.to_frac(prob)
    return dist((prob, 1 - prob), name=name)[0] \
        .with_output(name)


def _normalize_probs(probs):
    if len(probs) == 1:
        assert not isinstance(probs[0], int)

    probs = list(probs)
    if all(isinstance(x, int) for x in probs):
        total = sum(probs)
        probs = [(x, total) for x in probs]
    probs = fn.lmap(utils.to_frac, probs)

    if len(probs) == 1:
        probs = [probs[0], 1 - probs[0]]

    assert sum(probs) == 1
    return probs


def dist(probs, name=None, keep_seperate=False):
    """Distribution generated by mutually exclusive coins.

    Encoded using the common denominator method.
    """
    is_coin = len(probs) == 1
    probs = _normalize_probs(probs)
    bots = [p.denominator for p in probs]
    lcm = reduce(utils.lcm, bots, 1)
    word_len = max(math.ceil(math.log2(lcm)), 1)
    weights = map(lambda p: p.numerator*(lcm // p.denominator), probs)

    bits = atom(word_len, name, signed=False)
    if name is not None:
        bits = bits.with_output(name)

    name = bits.output

    const_true = ~(bits @ 0)
    total, coins, max_val = 0, [], 2**word_len
    for i, weight in enumerate(weights):
        lb = const_true if total == 0 else (bits >= total)
        total += weight
        ub = const_true if total == max_val else (bits < total)
        coins.append(lb & ub)

    is_valid = const_true if lcm == max_val else bits < lcm

    if keep_seperate:
        return coins, is_valid

    coins = reduce(UnsignedBVExpr.concat, coins) \
        .with_output(name)

    _dist = Distribution(expr=coins, valid=is_valid)
    return _dist[0] if is_coin else _dist


def binomial(n):
    chain = utils.chain(n)
    expr = UnsignedBVExpr(chain.unroll(n, only_last_outputs=True))
    const_true = ~(expr @ 0)
    return Distribution(expr=expr, valid=const_true)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['aiger_coins']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3,<20.0',
 'bidict>=0.18.3,<0.19.0',
 'funcy>=1.13,<2.0',
 'py-aiger-bv>=1.3.3,<2.0.0',
 'py-aiger>=3.6,<4.0']

extras_require = \
{'bdd': ['py-aiger-bdd>=0.2.6,<0.3.0'], 'sat': ['py-aiger-sat>=1.0,<2.0']}

setup_kwargs = {
    'name': 'py-aiger-coins',
    'version': '1.2.4',
    'description': 'Library for creating circuits that encode discrete distributions.',
    'long_description': '[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger-coins/status.svg)](https://cloud.drone.io/mvcisback/py-aiger-coins)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger-coins/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger-coins)\n[![Updates](https://pyup.io/repos/github/mvcisback/py-aiger-coins/shield.svg)](https://pyup.io/repos/github/mvcisback/py-aiger-coins/)\n\n[![PyPI version](https://badge.fury.io/py/py-aiger-coins.svg)](https://badge.fury.io/py/py-aiger-coins)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n\n# py-aiger-coins\nLibrary for creating circuits that encode discrete distributions and\nMarkov Decision Processes. The name comes from the random bit model of\ndrawing from discrete distributions using coin flips.\n\n<!-- markdown-toc start - Don\'t edit this section. Run M-x markdown-toc-refresh-toc -->\n**Table of Contents**\n\n- [py-aiger-coins](#py-aiger-coins)\n- [Install](#install)\n- [Usage](#usage)\n    - [Biased Coins](#biased-coins)\n    - [Distributions on discrete sets](#distributions-on-discrete-sets)\n    - [Distributions and Coins](#distributions-and-coins)\n        - [Manipulating Distributions](#manipulating-distributions)\n    - [Binomial Distributions](#binomial-distributions)\n    - [Markov Decision Processes and Probablistic Circuits](#markov-decision-processes-and-probablistic-circuits)\n\n<!-- markdown-toc end -->\n\n\n# Install\n\nTo install this library run:\n\n`$ pip install py-aiger-coins`\n\nNote that to actually compute probabilities, one needs to install with the bdd option.\n\n`$ pip install py-aiger-coins[bdd]`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then run:\n\n`$ poetry install`\n\n# Usage\n\nNote this tutorial assumes `py-aiger-bdd` has been installed (see the\nInstall section).\n\n## Biased Coins\n\nWe start by encoding a biased coin and computing its bias. The primary\nentrypoint for modeling coins is the `coin` function.\n\n```python\nfrom fractions import Fraction\n\nimport aiger_coins\n\nbias = Fraction(1, 3)\ncoin1 = aiger_coins.coin(bias)\ncoin2 = aiger_coins.coin((1, 3))  # or just use a tuple.\n\nassert coin1.prob() == coin2.prob() == prob\n```\n\n## Distributions on discrete sets\n\nWe now illustrate how to create a set of mutually exclusive coins that\nrepresent distribution over a finite set. For example, a biased three\nsided dice can be 1-hot encoded with:\n\n```python\ndice = aiger_coins.dist([1, 3, 2], name=\'x\')\ndice = aiger_coins.dist([(1, 6), (3, 6), (2, 6)], name=\'x\')  # equivalent\ndice = aiger_coins.dist([Fraction(1, 6), Fraction(3, 6), Fraction(2, 6)], name=\'x\')  # equivalent\n\nprint(dice.freqs())\n# (Fraction(1, 6), Fraction(1, 2), Fraction(1, 3))\n```\n\nLetting, `⚀ = dice[0]`, `⚁ = dice[1]`, `⚂ = dice[2]`, \n```python\none, two, three = dice[0], dice[1], dice[2]\n```\n\nWe can ask the probability of drawing an element of `{⚀, ⚁}` with:\n\n```python\nassert (one | two).prob() == Fraction(2, 3)\nassert (~three).prob() == Fraction(2, 3)\n```\n\n## Distributions and Coins\n\n`Distribution`s and `Coin`s are really just wrappers around two\n`aiger_bv.UnsignedBVExpr` objects stored in the `expr` and `valid`\nattributes.\n\nThe attributes `expr` and `valid` encode an expression over fair coin\nflips and which coin flips are "valid" respectively. Coins is a\nspecial type of `Distribution` where the expression is a predicate\n(e.g. has one output).\n\nNote that accessing the ith element of a `Distribution` results in a\n`Coin` encoding the probability of drawing that element.\n\n### Manipulating Distributions\n\nIn general `Distribution`s can me manipulated by manipulating the\n`.expr` attribution to reinterpret the coin flips or manipulating\n`.valid` to condition on different coin flip outcomes.\n\nOut of the box `Distribution`s support a small number of operations:\n`+, <, <=, >=, >, ==, !=, ~, |, &, ^, .concat`, which they inherit\nfrom `aiger_bv.UnsignedBVExpr`. When using the same `.valid` predicate\n(same coin flips), these operations only manipulate the `.expr`\nattribute.\n\nMore generally, one can use the `apply` method to apply an arbitrary\nfunction to the `.expr` attribute. For example, using the dice from\nbefore:\n\n```python\ndice2 = dice.apply(lambda expr: ~expr[2])\nassert dice2[0].freqs() == Fraction(2, 3)\n```\n\nOne can also change the assumptions made on the coin flips by using\nthe condition method, for example, suppose we condition on the coin\nflips never being all `False`. This changes the distribution\nas follows:\n\n```python\ncoins = dice.coins  #  Bitvector Expression of coin variables.\ndice3 = dice.condition(coins != 0)\n\nprint(dice3.freqs())\n# [Fraction(0, 5), Fraction(3, 5), Fraction(2, 5)]\n```\n\n## Binomial Distributions\n\nAs a convenience, `py-aiger-coins` also supports encoding Binomial\ndistributions.\n\n```python\nx = binomial(3)\n\nprint(x.freqs())\n# (Fraction(1, 8), Fraction(3, 8), Fraction(3, 8), Fraction(1, 8))\n```\n\n## Markov Decision Processes and Probablistic Circuits\n\n`aiger_coins` also supports modeling Probablistic Circuits, Markov\nDecision Process (MDPs), and Markov Chains (MDPs with no inputs).\n\nInternally, the `MDP` object is simply an `AIGBV` bitvector circuit\nwith some inputs annotated with distributions over their inputs.\n\nThe primary entropy point to modeling a Markov Decision Process is\nthe `circ2mdp` function.\n\n```python\nfrom aiger_bv import atom\nfrom aiger_coins import circ2mdp\n\nx = atom(3, \'x\', signed=False)\ny = atom(3, \'y\', signed=False)\nexpr = (x & y).with_output(\'x&y\')\n\nmdp1 = circ2mdp(expr)\nmdp1 = circ2mdp(expr.aigbv)  # equivalent\n```\n\n### Composition\n\n`MDP` can be composed using an API analogous to `aiger_bv.AIGBV` and\n`aiger.AIG` sequential circuits. In addition, `MDP` support being feed\nactions from a distribution via sequential composition.\n\n```python\n# Put a distribution over the y input.\ndist = aiger_coins.dist((0, 1, 2), name=\'y\')\n\nmdp2 = dist >> mdp1\nmdp2 = mdp1 << dist  # equivalent\nmdp2 = circ2mdp(expr, {\'y\': dist})  # equivalent\n\nassert mdp1.inputs == {\'x\', \'y\'}\nassert mdp2.inputs == {\'x\'}\n\nmdp3 = mdp2 | circ2mdp(aiger_bv.identity_gate(4, \'z\'))\nassert mdp3.inputs == {\'x\', \'z\'}\nassert mdp3.outputs == {\'x&y\', \'z\'}\n\nmdp4 = mdp3.feedback(inputs=[\'z\'], outputs=[\'x&y\'], keep_outputs=True)\nassert mdp4.inputs == {\'x\'}\nassert mdp4.outputs == {\'x&y\', \'z\'}\n```\n\n### Extracting Circuit\nOne can transform an `MDP` into an `AIG` or `AIGBV` object using\n`.aig` and `.aigbv` attributes. This adds as the coinflips explicitly\nas inputs and also adds a special output `##valid` that monitors if\nthe sequence of inputs and coin flips was valid.\n\n```python\nassert mdp.aigbv.outputs == mdp.outputs | {\'##valid\'}\n\nassert \'##valid[0]\' in mdp.aig.outputs\n```\n\n\n### Encoding and Decoding Traces\n\nOften times, one is interested in analyzing traces, sequences of\nstates and actions, through a Markov Decision Process. \n\nIn order to map this to an execution of an `MDP` object, one needs to\nfind a sequence of coin flip inputs such that feeding the actions and\nthe coin flip inputs into the circuit generated by `MDP.aigbv`.\n\nThis (and its inverse) can be done via the `MDP.encode_trc` and\n`MDP.decode_trc` methods.\n\nFor example, consider the simple MDP modeled by:\n\n```python\nfrom aiger_bv import atom\nfrom aiger_coins import circ2mdp\n\naction = atom(1, \'action\', signed=False)\nx_prev = atom(1, \'x_prev\', signed=False)\nc = atom(1, \'c\', signed=False)\n\nx_next = (x_prev & c & action).with_output(\'x_next\')\n\nsys = circ2mdp(x_next).feedback(\n    keep_outputs=True,\n    inputs=[\'x_prev\'], outputs=[\'x_next\'], initials=[(True,)],\n)\nsys <<= coin((1, 2), name=\'c\')\nassert sys.inputs == {\'action\'}\nassert sys.outputs == {\'x_next\'}\n```\n\nWe can encode and decode traces into this model as follows:\n\n```\n# Encoding and Decoding\n\nsys_actions = 3*[{\'action\': (True,)}]\nstates = 3*[{out: (True,)}]\n\nactions = sys.encode_trc(sys_actions, states)\nassert not any(v[\'c\'][0] for v in actions)\n\nsys_actions2, states2 = sys.decode_trc(actions)\nassert sys_actions2 == sys_actions\nassert states2 == states\n```\n',
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'url': 'https://github.com/mvcisback/py-aiger-coins',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import aiger_bv
import aiger_ptltl

import aiger_coins
from aiger_coins import coin
from aiger_coins.mdp import circ2mdp
from aiger_coins.utils import chain


def test_mdp_smoke():
    x = aiger_bv.identity_gate(2, 'x')
    y = aiger_bv.identity_gate(3, 'y')

    circ = x | y
    dist = aiger_coins.dist(((0, 3), (1, 3), (2, 3)), name='y')
    assert dist.expr.output == 'y'

    dyn = circ2mdp(circ, {'y': dist})
    assert dyn.inputs == {'x'}

    dyn2 = dist >> circ2mdp(circ)
    assert dyn2.inputs == {'x'}

    assert dyn2.aigbv.inputs == {'x'} | dist.inputs
    assert dyn2.aigbv.outputs == dyn2.outputs | {'##valid'}

    assert '##valid[0]' in dyn2.aig.outputs

    x = aiger_bv.atom(3, 'x', signed=False)
    y = aiger_bv.atom(3, 'y', signed=False)
    mdp = dist >> circ2mdp(x < y)
    assert mdp.inputs == {'x'}
    assert len(mdp.outputs) == 1


def test_closed_system():
    c1 = aiger_bv.atom(1, 'c1', signed=False)
    a = aiger_bv.atom(1, 'a', signed=False)

    dyn = circ2mdp(chain(n=4, state_name='s', action='a'))
    dyn <<= (c1 & a).with_output('a').aigbv
    dyn <<= coin((1, 8), name='c1')

    assert dyn.inputs == {'a'}
    assert dyn.outputs == {'s'}

    c2 = aiger_bv.atom(1, 'c2', signed=False)
    const_false = aiger_bv.atom(1, 0, signed=False)
    state = aiger_bv.atom(5, 's', signed=False)
    clip = state == 0b00001

    policy = circ2mdp(
        aiger_bv.ite(clip, const_false, c2).with_output('a')
    )
    policy <<= coin((1, 8), name='c2')

    sys = (policy >> dyn).feedback(
        inputs=['s'], outputs=['s'], latches=['s_prev2'], keep_outputs=True
    )
    assert sys.inputs == set()
    assert sys.outputs == {'s'}


def test_find_coin_flips():
    x, c = map(aiger_ptltl.atom, ('x', 'c'))

    sys = (x & c).historically().aig
    sys = circ2mdp(aiger_bv.aig2aigbv(sys))
    sys <<= coin((1, 2), name='c')

    assert sys.inputs == {'x'}
    assert len(sys.outputs) == 1

    out, *_ = sys.outputs
    sys_actions = 3*[{'x': (True,)}]
    states = 3*[{out: (True,)}]

    actions = sys.encode_trc(sys_actions, states)
    assert not any(v['c'][0] for v in actions)

    sys_actions2, states2 = sys.decode_trc(actions)
    assert sys_actions2 == sys_actions
    assert states2 == states


def test_mdp_readme():
    from aiger_bv import atom
    from aiger_coins import circ2mdp

    x = atom(3, 'x', signed=False)
    y = atom(3, 'y', signed=False)
    expr = (x & y).with_output('x&y')

    mdp1 = circ2mdp(expr)
    dist = aiger_coins.dist((0, 1, 2), name='y')

    mdp2 = dist >> mdp1

    assert mdp1.inputs == {'x', 'y'}
    assert mdp2.inputs == {'x'}

    mdp3 = mdp2 | circ2mdp(aiger_bv.identity_gate(3, 'z'))
    assert mdp3.inputs == {'x', 'z'}
    assert mdp3.outputs == {'x&y', 'z'}

    mdp4 = mdp3.feedback(inputs=['z'], outputs=['x&y'], keep_outputs=True)
    assert mdp4.inputs == {'x'}
    assert mdp4.outputs == {'x&y', 'z'}

    action = atom(1, 'action', signed=False)
    x_prev = atom(1, 'x_prev', signed=False)
    c = atom(1, 'c', signed=False)

    x_next = (x_prev & c & action).with_output('x_next')

    sys = circ2mdp(x_next).feedback(
        keep_outputs=True,
        inputs=['x_prev'], outputs=['x_next'], initials=[(True,)],
    )
    sys <<= coin((1, 2), name='c')
    assert sys.inputs == {'action'}
    assert sys.outputs == {'x_next'}

    sys_actions = 3*[{'action': (True,)}]
    states = 3*[{'x_next': (True,)}]

    actions = sys.encode_trc(sys_actions, states)
    assert not any(v['c'][0] for v in actions)

    sys_actions2, states2 = sys.decode_trc(actions)
    assert sys_actions2 == sys_actions
    assert states2 == states
